#include "machineReassignment.h"

void MachineReassignment::SpreadConstraint::reset(){
  nonzero.clear();
  zero.clear();
  for (unsigned int l = 0; l < con->mr->numLocations; ++l){
    counts[l] = 0;
    index[l] = l;
    zero.push_back(l);
  }
}

void MachineReassignment::SpreadConstraint::increaseCount(unsigned int location){
  if (++counts[location] == 1){
    unsigned int tail = *zero.rbegin();
    zero[index[location]] = tail;
    index[tail] = index[location];
    zero.pop_back();
    index[location] = nonzero.size();
    nonzero.push_back(location);
  }
}

void MachineReassignment::SpreadConstraint::decreaseCount(unsigned int location){
  if (--counts[location] == 0){
    unsigned int tail = *nonzero.rbegin();
    nonzero[index[location]] = tail;
    index[tail] = index[location];
    nonzero.pop_back();
    index[location] = zero.size();
    zero.push_back(location);
  }
}

MachineReassignment::int64 MachineReassignment::SpreadConstraint::calculatePenalty(Solution &solution){
  penalty = 0;
  reset();
  for (vector<unsigned int>::const_iterator ite = processes->begin(); ite != processes->end(); ++ite){
    increaseCount(con->mr->machines[solution.assignment[*ite]].location);
  }
  penalty = func(spreadMin - nonzero.size());
  return penalty;
}
void MachineReassignment::SpreadConstraint::update(unsigned int process, unsigned int machine0, unsigned int machine){
  unsigned int location_out = con->mr->machines[machine0].location;
  unsigned int location_in = con->mr->machines[machine].location;
  if (location_out != location_in){
    decreaseCount(location_out);
    increaseCount(location_in);
  }
}

MachineReassignment::int64 MachineReassignment::SpreadConstraint::updatePenalty(Solution &solution, Solver::Move *move){
  ppenalty = penalty;
  switch(move->type){
  case Solver::Shift:
    {
      Solver::ShiftMove *shift = dynamic_cast<Solver::ShiftMove*>(move);
      if (con->solver->processes[shift->process].service == service){
        update(shift->process, solution.assignment[shift->process], shift->machine);
      }
      break;
    }
  case Solver::Swap:
    {
      // Do Nothing 
      break;
    }
  }
  penalty = func(spreadMin - nonzero.size());
  return penalty - ppenalty;
}

void MachineReassignment::SpreadConstraint::calculateDiff(Solution &solution){
  // calculate diff; not necessary for swap
  vector<unsigned int>::const_iterator pite;
  vector<unsigned int>::iterator lite, mite;

  for (pite = processes->begin(); pite != processes->end(); ++pite){
    unsigned int process = *pite;
    unsigned int location = con->mr->machines[solution.assignment[process]].location;
    if (counts[location] > 1){
      if (penalty > 0){
        for (lite = zero.begin(); lite != zero.end(); ++lite){
          for (mite = con->mr->machinesInLocation[*lite].begin(); mite != con->mr->machinesInLocation[*lite].end(); ++mite){
            setShiftDiffPenalty(process, *mite, -1);
          }
        }
      }
    }
    else { // counts == 1
      if (nonzero.size() <= static_cast<unsigned>(spreadMin)){ 
        for (lite = nonzero.begin(); lite != nonzero.end(); ++lite){
          if (*lite != location){
            for (mite = con->mr->machinesInLocation[*lite].begin(); mite != con->mr->machinesInLocation[*lite].end(); ++mite){
              setShiftDiffPenalty(process, *mite, 1);
            }  
          }
        }  
      }  
    }  
  }  
}

void MachineReassignment::SpreadConstraint::prepare(Solver *solver){
  service = solver->mr->services[service0].id2;
  spreadMin = solver->services[service].spreadMin;
  if (spreadMin <= 1) return;

  con->addConstraint(this);
  counts.resize(solver->mr->numLocations);
  index.resize(solver->mr->numLocations);
  processes = &solver->services[service].processes;
  for (vector<unsigned int>::const_iterator ite = processes->begin(); ite != processes->end(); ++ite){
    involve(*ite, inf);
  }
}

// ----- Spread Constraints -----
void MachineReassignment::Constraints::initSpreadConstraints(){
  spreadConstraints.resize(mr->numServices);
  for (unsigned int s = 0; s < mr->numServices; ++s){
      spreadConstraints[s] = new SpreadConstraint(this, inf, s);
  }
}
void MachineReassignment::Constraints::prepareSpreadConstraints(){
  for (unsigned int s = 0; s < solver->numServices; ++s){
    spreadConstraints[solver->services[s].id]->prepare(solver);
  }
}
